#include "modelcanvas.h"
#include "video.h"
#include "animcontrol.h"
#include "gifexporter.h"

float animSpeed = 1.0f;
const float piover180 = 0.0174532925f;

IMPLEMENT_CLASS(ModelCanvas, wxGLCanvas)

BEGIN_EVENT_TABLE(ModelCanvas, wxGLCanvas)
    EVT_SIZE(ModelCanvas::OnSize)
    EVT_PAINT(ModelCanvas::OnPaint)
    EVT_ERASE_BACKGROUND(ModelCanvas::OnEraseBackground)
    EVT_MOUSE_EVENTS(ModelCanvas::OnMouse)
	EVT_KEY_DOWN(ModelCanvas::OnKey)

    EVT_TIMER(TIMER_ID, ModelCanvas::OnTimer)
	EVT_IDLE(ModelCanvas::OnIdle)

END_EVENT_TABLE()

#ifdef _WIN32
	// for timeGetTime:
	#pragma comment(lib,"Winmm.lib")

#else // for linux
	#include <sys/time.h>

	//typedef int DWORD;
	int timeGetTime()
	{
		static int start=0;
		static struct timeval t;
		gettimeofday(&t, NULL);
		if (start==0){
			start = t.tv_sec;
		}

		return (int)((t.tv_sec-start)*1000 + t.tv_usec/1000);
	}
#endif

DWORD pauseTime;

ModelCanvas::ModelCanvas(wxWindow *parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style, const wxString& name, int *attribList, const wxPalette& palette)
    : wxGLCanvas(parent, id, pos, size, style|wxFULL_REPAINT_ON_RESIZE, name, attribList, palette)
{
    init = false;
	
	srand(timeGetTime());

	time = 0;

	timer.SetOwner(this, TIMER_ID);
	timer.Start(TIME_STEP);
	//wxIdleEvent::SetMode(wxIDLE_PROCESS_SPECIFIED);

	fov = 45.0f;
	vRot = Vec3D(0.0f, 0.0f, 0.0f);
	vPos = Vec3D(0.0f, 0.0f, 5.0f);

	// Setup our 'default' lighting values.
	//Vec4D pos(0.8f, 0.2f, 0.5f, 0);
	lPos = Vec4D(-4.371f, 0.2f, 1.0f, 0.0f);
	lDif = Vec4D(1.0f, 1.0f, 1.0f, 1.0f);
	lAmb = Vec4D(0.2f, 0.2f, 0.2f, 0.0f);
	lightType = LT_DIRECTIONAL;

	// Setup our default colour values.
	bgcolor = Vec3D(0.1f, 0.2f, 0.4f);
	ltColor = Vec3D(1.0f, 1.0f, 1.0f);
	baseTrans = Vec3D(0.0f, 0.0f, 0.0f);

	model = 0;
	skyModel = 0;
	wmo = 0;
	animControl = 0;
	viewControl = 0;
	gifExporter = 0;
	modelType = MT_NORMAL;

	drawModel = true;
	drawParticles = true;
	drawMask = false;
	drawBones = false;
	drawBounds = false;
	drawLightDir = false;

	drawBackground = false;
	drawSky = false;
	bMouseLight = false;
	//useLights = true;
	useCamera = false;

	bSaving = false;
	
	root = new Attachment(NULL, NULL, -1, -1);
	sky = new Attachment(NULL, NULL, -1, -1);
}

ModelCanvas::~ModelCanvas()
{
	//clearAttachments();
	delete root;
	delete sky;
}


void ModelCanvas::OnSize(wxSizeEvent& event)
{
	wxGLCanvas::OnSize(event);

	if (init) 
		InitView();
}

void ModelCanvas::SetupProjection(float scale)
{
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	if (scale < 100.0f) {
		// "near" projection
	    gluPerspective(fov, aspect, 0.1f, 256.0f);
	} else {
		// "far" projection
	    gluPerspective(fov, aspect, 1.0f, 1024.0f);
	}
	glMatrixMode(GL_MODELVIEW);
}


void ModelCanvas::InitView()
{
	// set GL viewport
	int w, h;
	GetClientSize(&w, &h);
	SetCurrent();
	glViewport(0, 0, (GLint) w, (GLint) h);

	aspect = (float)w / (float)h;

	gXres = w;
	gYres = h;

	// update projection
	SetupProjection();
}

void ModelCanvas::OnEraseBackground(wxEraseEvent& WXUNUSED(event))
{
	// Do nothing
}

void ModelCanvas::ClearModel()
{
	if (model || wmo || root->model) {
		if (model)
			delete model;

		if (wmo)
			delete wmo;

		clearAttachments();

		model = NULL;
		wmo = NULL;
		root->model = NULL;
	}
}

Attachment* ModelCanvas::LoadModel(const char *fn)
{
	if (root->model != NULL) {
		clearAttachments();
		delete model;
		model = NULL;
		root->model = NULL;
	}

	if (wmo != NULL) {
		delete wmo;
		wmo = NULL;
	}

	model = new Model(fn, false);
	root->model = model;
	
	viewingModel = true;
	baseTrans = Vec3D(0.0f, 0.0f, 0.0f);

	return root;
}

Attachment* ModelCanvas::LoadCharModel(const char *fn)
{
	// Problem!  There seems to be 2 character models. 1 visible, 1 hidden (but there none the less)
	// Why are we doing it this way?  Is it to support mounts?  Just seems excess.
	
	//root->model = 0;
	return LoadModel(fn); //root->addChild(model, 0, -1);

	//ClearModel();
	//return root->addChild(fn, 0, -1);
}

void ModelCanvas::LoadWMO(std::string fn)
{
	if (root->model) 
		delete model;

	clearAttachments();
	model = 0;

	if (wmo) {
		if (wmo->name != fn) {
			delete wmo;
			wmo = 0;
		}
	}
	if (!wmo) {
		wmo = new WMO(fn);
		root->model = wmo;
	}
	viewingModel = false;
}


void ModelCanvas::clearAttachments()
{
	root->delChildren();
}

void ModelCanvas::OnMouse(wxMouseEvent& event)
{
	int px = event.GetX();
	int py = event.GetY();
	int pz = event.GetWheelRotation();

	// mul = multiplier in which to multiply everything to achieve a sense of control over the amount to move stuff by
	float mul = 1.0f;
	if (event.m_shiftDown)
		mul /= 10;
	if (event.m_controlDown)
		mul *= 10;
	if (event.m_altDown)
		mul *= 50;

	mul *= animSpeed;

	if (event.ButtonDown()) {
		mx = px;
		my = py;
		if (bMouseLight) // going to use vRot to hold our temp light position (technically, our lights rotation).
			vRot0 = Vec3D(lPos.x, lPos.y, lPos.z);
		else
			vRot0 = vRot;
		vPos0 = vPos;

	} else if (event.Dragging()) {
		int dx = mx - px;
		int dy = my - py;

		if (event.LeftIsDown()) {
			if (bMouseLight)
				return;

			vRot.x = vRot0.x - ((dy / 2.0f) * mul);
			vRot.y = vRot0.y - ((dx / 2.0f) * mul);

			viewControl->Refresh();

		} else if (event.RightIsDown()) {
			mul /= 100.0f;

			if (bMouseLight) {
				lPos.y = vRot0.y + dy*mul;
				lPos.x = vRot0.x - dx*mul;
			} else {
				vPos.x = vPos0.x - dx*mul;
				vPos.y = vPos0.y + dy*mul;

				viewControl->Refresh();
			}

		} else if (event.MiddleIsDown()) {
			if (!event.m_altDown) {
				mul = (mul / 20.0f) * dy;

				if (bMouseLight)
					lPos.z = vRot0.z - mul;
				else {
					Zoom(mul, false);
					my = py;
				}

			} else {
				mul = (mul / 1200.0f) * dy;
				Zoom(mul, true);
				my = py;
			}
		}

	} else if (event.GetEventType() == wxEVT_MOUSEWHEEL) {
		if (pz != 0) {
			mul = (mul / 120.0f) * pz;
			if (!wxGetKeyState(WXK_ALT)) {
				if (bMouseLight)
					lPos.z -= mul / 120.0f;
				else
					Zoom(mul, false);

			} else {
				mul /= 50.0f;
				Zoom(mul, true);
			}
		}
	} else if (event.GetEventType() == wxEVT_ENTER_WINDOW) {
		SetFocus();
	}
}

void ModelCanvas::InitGL()
{
	// init stuff
	InitExtensions();
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	/*
	if(supportAntiAlias)
		glEnable(GL_MULTISAMPLE_ARB);
	*/
	glShadeModel(GL_SMOOTH);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	glPixelStorei(GL_PACK_SWAP_BYTES, false);
	glPixelStorei(GL_PACK_LSB_FIRST, false);

	InitView();
	init = true;
	//drawBackground = false;
}

void ModelCanvas::OnPaint(wxPaintEvent& WXUNUSED(event))
{
	if (!init)
		InitGL();

// TODO: Fix for Linux
//#ifdef _WIN32
//	CheckKeyInput();
//#endif

	wxPaintDC dc(this);
	Render();

	if (bSaving == true) {
		gifExporter->SaveFrame();
		model->animManager->SetTimeDiff(gifExporter->TimeStep);
		model->animManager->Tick(gifExporter->TimeStep);
	}
}

inline void ModelCanvas::Render()
{
	//SetCurrent();

	glClearColor(bgcolor.x, bgcolor.y, bgcolor.z, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Ortho view, best way to understand it is to think of it as
	// being similar to DirectDraw
	// Draw the background image if any
	if(drawBackground && !drawMask)
	{
		glPushMatrix();

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		glMatrixMode(GL_MODELVIEW);					// Select The Modelview Matrix
		glLoadIdentity();

		glOrtho(0, gXres, 0, gYres, -1, 1);
		glEnable(GL_TEXTURE_2D);				// Enable 2D Texture Mapping
		glDisable(GL_DEPTH_TEST);
		
		glBindTexture(GL_TEXTURE_2D, bgTexture);
		glBegin(GL_QUADS);
			glTexCoord2f(0.0f, 0.0f); glVertex2i(0, 0);
			glTexCoord2f(1.0f, 0.0f); glVertex2i(gXres, 0);
			glTexCoord2f(1.0f, 1.0f); glVertex2i(gXres, gYres);
			glTexCoord2f(0.0f, 1.0f); glVertex2i(0,  gYres);
		glEnd();

		glPopMatrix();								// Restore The Old Modelview Matrix
	}

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	//if (root->model) {
		if (model) 
			model->animcalc = false;		
		
		// All our lighting related rendering code
		bool lit = false;
		Vec4D la;

		// Use model lighting?
		if (model && (lightType == LT_MODEL_ONLY)) {
			if (model->header.nLights > 0) {
				la = Vec4D(0.0f, 0.0f, 0.0f, 1.0f);
				lit = true;
			} else {
				la = Vec4D(1.0f, 1.0f, 1.0f, 1.0f);
			}
		} else {
			la = Vec4D(0.5f, 0.5f, 0.5f, 1.0f);
		}

		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, la);

		glEnable(GL_COLOR_MATERIAL);
		glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
		glColor4f(ltColor.x, ltColor.y, ltColor.z, 1.0f);

		// Why are we using all 8 lights?  
		// In fact, why are we setting all these settings then disabling the light?
		for (int i=0; i<8; i++) {
			GLuint light = GL_LIGHT0 + i;
			//GLuint light = GL_LIGHT4;
			glLightf(light, GL_CONSTANT_ATTENUATION, 0.0f);
			glLightf(light, GL_LINEAR_ATTENUATION, 0.7f);
			glLightf(light, GL_QUADRATIC_ATTENUATION, 0.03f);
			//glDisable(light);
		}

		if (!lit) {
			if (lightType == LT_DIRECTIONAL) {
				glLightfv(GL_LIGHT0, GL_POSITION, lPos);
				//glLightfv(GL_LIGHT0, GL_DIFFUSE, lDif);
				glLightfv(GL_LIGHT0, GL_DIFFUSE, ltColor);

				lAmb = Vec4D(ltColor.x / 5, ltColor.y / 5, ltColor.z / 5, 0.0f); // To keep the ambience lighting in the correct proportion.
				glLightfv(GL_LIGHT0, GL_AMBIENT, lAmb);

				glEnable(GL_LIGHT0);
			} else if (lightType==LT_AMBIENT) {
				glLightModelfv(GL_LIGHT_MODEL_AMBIENT, Vec4D(ltColor.x, ltColor.y, ltColor.z, 1.0f));
			}
		}

		// render the skybox, if any
		if (drawSky && skyModel && sky->model) {
			glPushMatrix();
			
			float fScale = 100.0f / skyModel->rad;
			
			glTranslatef(0.0f, 0.0f, -5.0f);
			glScalef(fScale, fScale, fScale);
			sky->draw(this);

			glPopMatrix();
		}

		// setup the view/projection
		if (model && useCamera && model->hasCamera) {
			model->cam.setup();
		} else {

			// update projection
			SetupProjection(modelsize);

			// update view
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			
			glTranslatef(vPos.x, vPos.y, -vPos.z);
			glRotatef(vRot.x, 1.0f, 0.0f, 0.0f);
			glRotatef(vRot.y, 0.0f, 1.0f, 0.0f);
			glRotatef(vRot.z, 0.0f, 0.0f, 1.0f);
			glTranslatef(baseTrans.x, baseTrans.y, baseTrans.z);
		}
		
		// render our main model
		if (model || wmo) {		
			glDisable(GL_CULL_FACE);
			root->draw(this);

			// render our particles, we do this afterwards so that all the particles display "OK" without having things like shields "overwriting" the particles.
			if (drawParticles && !drawMask && modelType!=MT_WMO) {
				glEnable(GL_LIGHTING);
				glEnable(GL_TEXTURE_2D);
				//glDisable(GL_CULL_FACE);
				glDepthMask(GL_FALSE);

				root->drawParticles(this);

				glDepthMask(GL_TRUE);
			}
		}

		// Draws our "lighting direction guide-line"
		if (drawLightDir) {
			glRotatef(vRot.y, 0.0f, -1.0f, 0.0f);
			glBegin(GL_LINES);
				glVertex3f(0.0f, 0.0f, 0.0f);
				glVertex3f(lPos.x, lPos.y, lPos.z);
			glEnd();
		}
		
		
	//}
	

	glFlush();
    SwapBuffers();
}

void Attachment::draw(ModelCanvas *c)
{
	if (!c)
		return;

	glPushMatrix();

	if (model) {
		model->reset();
		setup();

		// no need to scale if its already 100%
		if (scale != 1.0f)
			glScalef(scale, scale, scale);
		
		if (c->drawModel) {
			if(c->drawMask) {
				glDisable(GL_TEXTURE_2D);
				glDisable(GL_LIGHTING);
				glDisable(GL_DEPTH_TEST);
			} else {
				glEnable(GL_LIGHTING);
				glEnable(GL_TEXTURE_2D);

				glEnable(GL_DEPTH_TEST);
				//glDepthFunc(GL_LEQUAL);
			}

			model->draw();
			
			// Disable lighting and textures if we haven't already
			if(!c->drawMask) {
				glDisable(GL_LIGHTING);
				glDisable(GL_TEXTURE_2D);
			}

			if (c->viewingModel) {
				Model *m = static_cast<Model*>(model);
				if (!m)
					return;

				if (!c->drawModel) {
					m->trans = 0.0f;
					m->draw();
					m->trans = 1.0f;
				}

				if (c->drawBounds) 
					m->drawBoundingVolume();
				
				if (c->drawBones) 
					m->drawBones();
			}
		}
	}

	// children:
	for (size_t i=0; i<children.size(); i++) {
		children[i]->draw(c);
	}

	glPopMatrix();
}

void Attachment::drawParticles(ModelCanvas *c)
{
	if (!c)
		return;

	glPushMatrix();

	if (model) {
		Model *m = static_cast<Model*>(model);
		if (!m)
			return;
		
		model->reset();
		setup();

		if (c->modelType!=MT_WMO && m->hasParticles) 
				m->drawParticles();
	}
	
	// children:
	for (size_t i=0; i<children.size(); i++)
		children[i]->drawParticles(c);

	glPopMatrix();
}



void Attachment::tick(float dt)
{
	if (model)
		model->update(dt);
	for (size_t i=0; i<children.size(); i++)
		children[i]->tick(dt);
}


void ModelCanvas::OnTimer(wxTimerEvent& event)
{
	if (bSaving == false) {
		tick();
		Refresh(false);
	}
}

void ModelCanvas::OnIdle(wxIdleEvent& event)
{
	if (bSaving == true) {
		tick();
		Refresh(false);
	}
}

void ModelCanvas::tick()
{
	static DWORD lastTime = timeGetTime();
	int ddt = 0;

	// Time stuff
	//time = float();
	ddt = (timeGetTime() - lastTime) * animSpeed;
	lastTime = timeGetTime();
	// --

	AnimManager *animManager = NULL;

	globalTime += (ddt);

	if (model) {
		if (model->animManager) {
			animManager = model->animManager;
			if (animManager->IsPaused())
				ddt = 0;
			else if (!animManager->IsParticlePaused())
				ddt = animManager->GetTimeDiff();
		}
		
		//animManager->Tick(ddt);
		root->tick(ddt);
		//root->tick(ddt/1000.0f);
	//} else if (wmo) {
	//	root->tick(ddt/1000.0f);
	}

	if (drawSky && skyModel && skyModel->animManager) {
		animManager = skyModel->animManager;
		
		//animManager->Tick(ddt);
		sky->tick(ddt);
	}

	
}

/*
void ModelCanvas::TogglePause()
{
	if (!bPaused) {
		// pause
		bPaused = true;
		pauseTime = timeGetTime();

	} else {
		// unpause
		DWORD t = timeGetTime();
		deltaTime += t - pauseTime;
		if (time==0) deltaTime = t;
		bPaused = false;
	}
}
*/

void ModelCanvas::ResetView()
{
	vRot = Vec3D(0,-90,0);
	vPos = Vec3D(0, 0, 5);

	bool isSkyBox = (model->name.substr(0,3)=="Env");
	if (!isSkyBox) {
		if (model->name.find("SkyBox")<model->name.length()) isSkyBox = true;
	}

	if (isSkyBox) {
		// for skyboxes, don't zoom out ;)
		vPos.y = vPos.z = 0.0f;
	} else {
		vPos.z = model->rad * 2.0f;
		if (vPos.z < 3.0f) vPos.z = 3.0f;
		if (vPos.z > 500.0f) vPos.z = 500.0f;
		
		//ofsy = (model->anims[model->currentAnim].boxA.y + model->anims[model->currentAnim].boxB.y) * 0.5f;
		vPos.y = -model->rad * 0.5f;
		if (vPos.y > 50) vPos.y = 50;
		if (vPos.y < -50) vPos.y = -50;
	}

	modelsize = model->rad * 2.0f;
	
	if (model->name.substr(0,4)=="Item") vRot.y = 0; // items look better facing right by default
}

void ModelCanvas::ResetViewWMO(int id)
{
	if (!wmo || id>=wmo->nGroups) return;

	vRot = Vec3D(0.0f, -90.0f, 0.0f);
	vPos = Vec3D(0.0f, 0.0f, 5.0f);
	Vec3D mid;

	if (id==-1) {
		vPos.z = (wmo->v2-wmo->v1).length();
		mid = (wmo->v1+wmo->v2)*0.5f;
	} else {
		// zoom/center on current WMO group
		WMOGroup &g = wmo->groups[id];
		vPos.z = (g.v2-g.v1).length();
		mid = (g.v1+g.v2)*0.5f;
	}

	modelsize = vPos.z;

	baseTrans = Vec3D(mid.x,mid.z,-mid.y) * -1.0f;
	if (vPos.z < 3.0f) vPos.z = 3.0f;
	if (vPos.z > 500.0f) vPos.z = 500.0f;
}

void ModelCanvas::loadBackground(wxString filename)
{
	FILE *file = NULL;
	BYTE *buffer = NULL;
	//GLint format = GL_RGBA;
	CxImage *image = NULL;

	if (!wxFile::Exists(filename))
		return;

	file = fopen(filename.fn_str(), "rb");
	if (!file) {
		fclose(file);
		return;
	}

	bgImagePath = filename;

	// Setup the OpenGL Texture stuff
	glGenTextures(1, &bgTexture);
	glBindTexture(GL_TEXTURE_2D, bgTexture);
	
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);	// Linear Filtering
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);	// Linear Filtering

	// Get the file extension and load the file
	wxString tmp = filename.Mid(filename.Length() - 3, 3);
	tmp.LowerCase();

	if (tmp == "bmp")
		image = new CxImage(file, CXIMAGE_FORMAT_BMP);
	else if (tmp == "tga")
		image = new CxImage(file, CXIMAGE_FORMAT_TGA);
	else if (tmp == "jpg")
		image = new CxImage(file, CXIMAGE_FORMAT_JPG);
	else {
		fclose(file);
		return;
	}

	if (image == NULL) {
		fclose(file);
		return;
	}

	// TODO: Check the size of the images, resize as needed?
	bool textureXSafe = false;
	bool textureYSafe = false;
	for (int i=2; i < 2100; i+=i) {
		if (image->GetWidth() == i)
			textureXSafe = true;
		if (image->GetHeight() == i)
			textureYSafe = true;
	}

	if (!textureXSafe || !textureYSafe) {
		wxMessageBox(_T("The dimensions of that image are not to a power of 2.  eg. 64x64, 128x128, 256x256, 512x512\nThis may result in unstable results, resizing to 256x256"), _T("Caution!"));
		if (image->GetBpp() < 24)
			image->IncreaseBpp(24);
		image->Resample(256,256);
	}

	long size = image->GetWidth() * image->GetHeight() * 4;
	image->Encode2RGBA(buffer, size);
	
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, image->GetWidth(), image->GetHeight(), 0, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
	drawBackground = true;

	fclose(file);
	delete image;
	delete buffer;
}

void ModelCanvas::Zoom(float f, bool rel)
{
	if (rel) {
		float cosx = cos(vRot.x * piover180);
		vPos.x += cos(vRot.y * piover180) * cosx * f;
		vPos.y += sin(vRot.x * piover180) * sin(vRot.y * piover180) * f;
		vPos.z += sin(vRot.y * piover180) * cosx * f;
	} else {
		vPos.z -= f;
	}

	viewControl->Refresh();
}

// Check for keyboard input
void ModelCanvas::OnKey(wxKeyEvent &event)
{
	// error checking
	if(!model) 
		return;

	// Make sure its the canvas that has focus before continuing
	/*
	wxWindow *win = wxWindow::FindFocus();
	if(!win)
		return;
	wxGLCanvas *gl = wxDynamicCast(win, wxGLCanvas);
	if(!gl)
		return;
	*/
	// --
	int keycode = event.GetKeyCode(); 

	// animation speed toggles
	if (keycode == '0')	animControl->SetAnimSpeed(1.0f);
	else if (keycode == '1')	animControl->SetAnimSpeed(0.1f);
	else if (keycode == '2')	animControl->SetAnimSpeed(0.2f);
	else if (keycode == '3')	animControl->SetAnimSpeed(0.3f);
	else if (keycode == '4')	animControl->SetAnimSpeed(0.4f);
	else if (keycode == '5')	animControl->SetAnimSpeed(0.5f);
	else if (keycode == '6')	animControl->SetAnimSpeed(0.6f);
	else if (keycode == '7')	animControl->SetAnimSpeed(0.7f);
	else if (keycode == '8')	animControl->SetAnimSpeed(0.8f);
	else if (keycode == '9')	animControl->SetAnimSpeed(0.9f);
	// --

	// Turning
	if (wxGetKeyState(WXK_LEFT)) {
		vRot.y += (2.4f * animSpeed);
		viewControl->Refresh();

		if (vRot.y > 360) vRot.y -= 360;
		if (vRot.y < 0) vRot.y += 360;

	} else if (wxGetKeyState(WXK_RIGHT)) {
		vRot.y -= (2.4f * animSpeed);
		viewControl->Refresh();

		if (vRot.y > 360) vRot.y -= 360;
		if (vRot.y < 0) vRot.y += 360;
	}
	// --

	// Moving forward/backward
	float speed = 0.0f;
	if ((modelType!=MT_WMO) && (model->animated==true))
		speed = (model->anims[model->currentAnim].moveSpeed / 80.0f) * animSpeed;
	else
		speed = 2.0f * animSpeed;

	if (wxGetKeyState(WXK_UP))
		Zoom(speed, true);
	else if (wxGetKeyState(WXK_DOWN))
		Zoom(-speed, true);
	// --
	
}

/*
void ModelCanvas::SetFrame(int frame)
{
	// 
}
*/

void Attachment::setup()
{
	if (parent==0) 
		return;
	if (parent->model) 
		parent->model->setupAtt(id);
}

Attachment* Attachment::addChild(const char *modelfn, int id, int slot, float scale)
{
	if (!modelfn || !strlen(modelfn) || id<0) return 0;
	Model *m = new Model(modelfn, true);
	if (m && m->ok) {
		return addChild(m, id, slot, scale);
	} else {
		delete m;
		return 0;
	}
}

Attachment* Attachment::addChild(Displayable *disp, int id, int slot, float scale)
{
	Attachment *att = new Attachment(this, disp, id, slot, scale);
	children.push_back(att);
	return att;
}

void Attachment::delChildren()
{
	for (size_t i=0; i<children.size(); i++) delete children[i];
	children.clear();
}

void Attachment::delSlot(int slot)
{
	for (size_t i=0; i<children.size(); ) {
		if (children[i]->slot == slot) {
			delete children[i];
			children.erase(children.begin() + i);
		} else i++;
	}
}

